#include "config.h"

#include "dir-menu.h"

#include "file-menu-item.h"

#include <string.h>

enum {
	BE_DIR_MENU_PROPERTY_URI         = 1,
	BE_DIR_MENU_PROPERTY_ICON_LOADER,
	BE_DIR_MENU_PROPERTY_ICON_SIZE
};

static GtkWidget *be_dir_menu_item_new(BeDirMenu *menu, const gchar *child) {
	gchar     *uri  = g_build_path("/", menu->uri, child, NULL);
	GtkWidget *item = be_file_menu_item_new(uri, menu->icon_loader, menu->icon_size);

	g_free(uri);

	return item;
}

static guint be_dir_menu_item_find_insertion_index(BeDirMenu *menu, BeFileMenuItem *item) {
	const gchar *name = be_file_menu_item_get_name(item);
	GList *cur;
	guint a;

	for(a = 0, cur = gtk_container_get_children(GTK_CONTAINER(menu)); cur; ++a, cur = g_list_next(cur)) {
		const gchar *cur_name = be_file_menu_item_get_name(BE_FILE_MENU_ITEM(cur->data));

		if(strcmp(name, cur_name) < 0)
			return a;
	}

	return a;
}

static void be_dir_menu_append(BeDirMenu *menu, const gchar *child) {
	if(child && *child && child[0] != '.' && strcmp(child, ".directory") != 0) {
		GtkWidget *item = be_dir_menu_item_new(menu, child);

		if(item) {
			guint index = be_dir_menu_item_find_insertion_index(menu, BE_FILE_MENU_ITEM(item));

			g_hash_table_insert(menu->children, g_strdup(child), item);

			gtk_widget_show(item);
			gtk_menu_shell_insert(GTK_MENU_SHELL(menu), item, index);
		}
	}
}

static void be_dir_menu_fill(BeDirMenu *menu) {
	GnomeVFSDirectoryHandle *dir;

	if(gnome_vfs_directory_open(&dir, menu->uri, GNOME_VFS_FILE_INFO_DEFAULT) != GNOME_VFS_OK) {
		g_warning("could not open directory \"%s\"", menu->uri);
	} else {
		GnomeVFSFileInfo *file = gnome_vfs_file_info_new();

		while(gnome_vfs_directory_read_next(dir, file) == GNOME_VFS_OK) {
			be_dir_menu_append(menu, file->name);

			gnome_vfs_file_info_clear(file);
		}

		gnome_vfs_file_info_unref(file);

		gnome_vfs_directory_close(dir);
	}
}
static gboolean be_dir_menu_clear_hrfunc(gpointer key, gpointer value, gpointer user_data) {
	BeDirMenu *dir_menu = user_data;

	gtk_container_remove(GTK_CONTAINER(dir_menu), GTK_WIDGET(value));

	return TRUE;
}
static void be_dir_menu_clear(BeDirMenu *menu) {
	g_hash_table_foreach_remove(menu->children, &be_dir_menu_clear_hrfunc, menu);
}
static void be_dir_menu_reset(BeDirMenu *menu) {
	be_dir_menu_clear(menu);
	be_dir_menu_fill(menu);
}

static void be_dir_menu_show_idle_hfunc(const gchar *key, GtkWidget *value, gpointer user_data) {
	be_readyable_ready(BE_READYABLE(value));
}
static gboolean be_dir_menu_show_idle(BeDirMenu *dir_menu) {
	g_hash_table_foreach(dir_menu->children, (GHFunc) &be_dir_menu_show_idle_hfunc, NULL);
	dir_menu->idle = 0;
	return FALSE;
}
static void be_dir_menu_show(GtkWidget *widget, gpointer user_data) {
	BeDirMenu *dir_menu = BE_DIR_MENU(widget);

	be_readyable_ready(BE_READYABLE(dir_menu));

	if(dir_menu->idle == 0)
		dir_menu->idle = g_idle_add((GSourceFunc) &be_dir_menu_show_idle, dir_menu);
}
static void be_dir_menu_hide(GtkWidget *widget, gpointer user_data) {
	BeDirMenu *dir_menu = BE_DIR_MENU(widget);

	if(dir_menu->idle != 0)
		g_idle_remove_by_data(dir_menu);

	be_readyable_unready(BE_READYABLE(dir_menu));
}

static void be_dir_menu_instance_init(GTypeInstance *instance, gpointer g_class) {
	BeDirMenu *self = (BeDirMenu *) instance;

	/* Do stuff */
	self->icon_loader = NULL;
	self->icon_size   = GTK_ICON_SIZE_MENU;
	self->ready       = 0;
	self->uri         = NULL;
	self->children    = g_hash_table_new_full(g_str_hash, g_str_equal, g_free, NULL);
	self->idle        = 0;

	g_signal_connect(self, "show", G_CALLBACK(&be_dir_menu_show), NULL);
	g_signal_connect(self, "hide", G_CALLBACK(&be_dir_menu_hide), NULL);
}

static void be_dir_menu_set_property(GObject *object, guint property_id, const GValue *value, GParamSpec *pspec) {
	BeDirMenu *self = BE_DIR_MENU(object);

	switch(property_id) {
		case BE_DIR_MENU_PROPERTY_URI:
			be_dir_menu_set_uri(self, g_value_get_string(value));
			break;
		case BE_DIR_MENU_PROPERTY_ICON_LOADER:
			be_dir_menu_set_icon_loader(self, BE_ICON_LOADER(g_value_get_object(value)));
			break;
		case BE_DIR_MENU_PROPERTY_ICON_SIZE:
			be_dir_menu_set_icon_size(self, g_value_get_uint(value));
			break;
		default:
			g_assert_not_reached();
			break;
	}
}
static void be_dir_menu_get_property(GObject *object, guint property_id, GValue *value, GParamSpec *pspec) {
	BeDirMenu *self = BE_DIR_MENU(object);

	switch(property_id) {
		case BE_DIR_MENU_PROPERTY_URI:
			g_value_set_string(value, self->uri);
			break;
		case BE_DIR_MENU_PROPERTY_ICON_LOADER:
			g_value_set_object(value, self->icon_loader);
			break;
		case BE_DIR_MENU_PROPERTY_ICON_SIZE:
			g_value_set_uint(value, self->icon_size);
			break;
		default:
			g_assert_not_reached();
			break;
	}
}

static GObject *be_dir_menu_constructor(GType type, guint n_construct_properties, GObjectConstructParam *construct_properties) {
	GObject *obj;

	{ /* Invoke parent constructor. */
		BeDirMenuClass *klass        = BE_DIR_MENU_CLASS(g_type_class_peek(BE_DIR_MENU_TYPE));
		GObjectClass   *parent_class = G_OBJECT_CLASS(g_type_class_peek_parent(klass));

		obj = parent_class->constructor(type, n_construct_properties, construct_properties);
	}

	/* Do stuff */

	return obj;
}

#if 0
static void be_dir_menu_dispose(GObject *object) {
	BeDirMenuClass *klass        = BE_DIR_MENU_CLASS(g_type_class_peek(BE_DIR_MENU_TYPE));
	GObjectClass   *parent_class = G_OBJECT_CLASS(g_type_class_peek_parent(klass));

	/* Do stuff */

	/* Invoke parent dispose */
	parent_class->dispose(object);
}
#endif

static void be_dir_menu_finalize(GObject *object) {
	BeDirMenu      *self         = BE_DIR_MENU(object);
	BeDirMenuClass *klass        = BE_DIR_MENU_CLASS(g_type_class_peek(BE_DIR_MENU_TYPE));
	GObjectClass   *parent_class = G_OBJECT_CLASS(g_type_class_peek_parent(klass));

	/* Do stuff */
	g_free(self->uri);
	g_hash_table_destroy(self->children);
	if(self->idle != 0)
		g_idle_remove_by_data(self);

	/* Invoke parent finalize */
	parent_class->finalize(object);
}

static void be_dir_menu_class_init(gpointer g_class, gpointer g_class_data) {
	GObjectClass   *gobject_class = G_OBJECT_CLASS(g_class);
/*	BeDirMenuClass *klass         = BE_DIR_MENU_CLASS(g_class); */
	GParamSpec     *pspec;

	gobject_class->constructor  = &be_dir_menu_constructor;
/*	gobject_class->dispose      = &be_dir_menu_dispose; */
	gobject_class->finalize     = &be_dir_menu_finalize;
	gobject_class->set_property = &be_dir_menu_set_property;
	gobject_class->get_property = &be_dir_menu_get_property;

	pspec = g_param_spec_string("uri",
		"URI",
		"URI of the directory to display in the menu",
		NULL,
		G_PARAM_READWRITE);
	g_object_class_install_property(gobject_class,
		BE_DIR_MENU_PROPERTY_URI,
		pspec);

	pspec = g_param_spec_object("icon-loader",
		"Icon loader to pass to items",
		"Icon loader to pass to items",
		G_TYPE_OBJECT,
		G_PARAM_READWRITE);
	g_object_class_install_property(gobject_class,
		BE_DIR_MENU_PROPERTY_ICON_LOADER,
		pspec);

	pspec = g_param_spec_uint("icon-size",
		"Icon size to pass to items",
		"Icon size to pass to items",
		0,
		~0U,
		GTK_ICON_SIZE_MENU,
		G_PARAM_READWRITE);
	g_object_class_install_property(gobject_class,
		BE_DIR_MENU_PROPERTY_ICON_SIZE,
		pspec);
}

static void be_dir_menu_ready(BeReadyable *readyable) {
	BeDirMenu *dir_menu = BE_DIR_MENU(readyable);

	if(dir_menu->ready == 0)
		be_dir_menu_reset(dir_menu);

	++dir_menu->ready;
}
static void be_dir_menu_unready(BeReadyable *readyable) {
	BeDirMenu *dir_menu = BE_DIR_MENU(readyable);

	--dir_menu->ready;

	if(dir_menu->ready == 0)
		be_dir_menu_clear(dir_menu);
}

static void be_dir_menu_readyable_interface_init(gpointer g_iface, gpointer iface_data) {
	BeReadyableClass *klass = (BeReadyableClass *) g_iface;

	klass->ready   = &be_dir_menu_ready;
	klass->unready = &be_dir_menu_unready;
}

GType be_dir_menu_get_type(void) {
	static GType type = 0;

	if(type == 0) {
		static const GTypeInfo info = {
			sizeof(BeDirMenuClass),     /* class_size */
			NULL,                       /* base_init */
			NULL,                       /* base_finalize */
			&be_dir_menu_class_init,    /* class_init */
			NULL,                       /* class_destroy */
			NULL,                       /* class_data */
			sizeof(BeDirMenu),          /* instance_size */
			0,                          /* n_preallocs */
			&be_dir_menu_instance_init, /* instance_init */
			NULL                        /* value_table */
		};
		static const GInterfaceInfo readyable_info = {
			(GInterfaceInitFunc) &be_dir_menu_readyable_interface_init, /* interface_init */
			NULL,                                                       /* interface_finalize */
			NULL                                                        /* interface_data */
		};

		type = g_type_register_static(GTK_TYPE_MENU, "BeDirMenuType", &info, 0);

		g_type_add_interface_static(type, BE_READYABLE_TYPE, &readyable_info);
	}

	return type;
}

GtkWidget *be_dir_menu_new(const gchar *uri, BeIconLoader *icon_loader, GtkIconSize icon_size) {
	GtkWidget *dir_menu = gtk_widget_new(BE_DIR_MENU_TYPE,
		"uri", uri,
		"icon-loader", icon_loader,
		"icon-size", icon_size,
		NULL);

	return dir_menu;
}

void be_dir_menu_set_uri(BeDirMenu *dir_menu, const gchar *uri) {
	g_free(dir_menu->uri);
	dir_menu->uri = g_strdup(uri);

	be_dir_menu_clear(dir_menu);

	g_object_notify(G_OBJECT(dir_menu), "uri");
}

void be_dir_menu_set_icon_loader(BeDirMenu *dir_menu, BeIconLoader *icon_loader) {
	if(dir_menu->icon_loader)
		g_object_unref(G_OBJECT(dir_menu->icon_loader));

	dir_menu->icon_loader = icon_loader;

	if(dir_menu->icon_loader) {
		GList *cur;

		g_object_ref(G_OBJECT(dir_menu->icon_loader));

		for(cur = gtk_container_get_children(GTK_CONTAINER(dir_menu)); cur; cur = g_list_next(cur))
			be_file_menu_item_set_icon_loader(BE_FILE_MENU_ITEM(cur->data), dir_menu->icon_loader);
	}

	g_object_notify(G_OBJECT(dir_menu), "icon-loader");
}

void be_dir_menu_set_icon_size(BeDirMenu *dir_menu, GtkIconSize icon_size) {
	if(dir_menu->icon_size != icon_size) {
		GList *cur;

		dir_menu->icon_size = icon_size;

		for(cur = gtk_container_get_children(GTK_CONTAINER(dir_menu)); cur; cur = g_list_next(cur))
			be_file_menu_item_set_icon_size(BE_FILE_MENU_ITEM(cur->data), dir_menu->icon_size);

		g_object_notify(G_OBJECT(dir_menu), "icon-size");
	}
}
