#include "config.h"

#include "icon-loader.h"

/* Thanks to gnome-panel once again. */

#include <libgnome/gnome-desktop-item.h>

#include <string.h>

/* BeIconSized, used by BeIcon for a single size */
typedef struct {
	GtkIconSize  icon_size;
	GdkPixbuf   *pixbuf;
	GList       *images;
} BeIconSized;

static void be_icon_sized_image_weak_notify(BeIconSized *icon_sized, GObject *former_object) {
	icon_sized->images = g_list_remove(icon_sized->images, former_object);
}

static BeIconSized *be_icon_sized_new(GtkIconSize icon_size) {
	BeIconSized *icon_sized = g_new0(BeIconSized, 1);

	icon_sized->icon_size = icon_size;

	return icon_sized;
}
static void be_icon_sized_free(BeIconSized *icon_sized) {
	if(icon_sized->pixbuf)
		g_object_unref(icon_sized->pixbuf);

	if(icon_sized->images) {
		GList *cur;

		for(cur = icon_sized->images; cur; cur = g_list_next(cur))
			g_object_weak_unref(G_OBJECT(cur), (GWeakNotify) &be_icon_sized_image_weak_notify, icon_sized);

		g_list_free(icon_sized->images);
	}
	
	g_list_free(icon_sized->images);

	g_free(icon_sized);
}

static void be_icon_sized_set(BeIconSized *icon_sized, GdkPixbuf *pixbuf) {
	GList *cur;

	if(icon_sized->pixbuf)
		g_object_unref(icon_sized->pixbuf);

	icon_sized->pixbuf = pixbuf;

	if(icon_sized->pixbuf) {
		g_object_ref(icon_sized->pixbuf);

		for(cur = icon_sized->images; cur; cur = g_list_next(cur))
			gtk_image_set_from_pixbuf(GTK_IMAGE(cur->data), icon_sized->pixbuf);
	} else {
		for(cur = icon_sized->images; cur; cur = g_list_next(cur))
			gtk_image_set_from_stock(GTK_IMAGE(cur->data), GTK_STOCK_MISSING_IMAGE, icon_sized->icon_size);
	}
}

static void be_icon_sized_add_image(BeIconSized *icon_sized, GtkImage *image) {
	icon_sized->images = g_list_prepend(icon_sized->images, image);

	g_object_weak_ref(G_OBJECT(image), (GWeakNotify) &be_icon_sized_image_weak_notify, icon_sized);

	if(icon_sized->pixbuf)
		gtk_image_set_from_pixbuf(image, icon_sized->pixbuf);
	else
		gtk_image_set_from_stock(image, GTK_STOCK_MISSING_IMAGE, icon_sized->icon_size);
}

/* BeIcon caches pixbufs for icons of varying sizes */
typedef struct {
	gchar        *name;
	gchar        *path;
	gint          largest;
	GtkIconTheme *icon_theme;
	GdkPixbuf    *pixbuf;
	GHashTable   *sized;
} BeIcon;

static GdkPixbuf *be_icon_load_pixbuf(BeIcon *icon) {
	GdkPixbuf *pixbuf = gdk_pixbuf_new_from_file_at_size(icon->path, icon->largest, icon->largest, NULL);

	if(pixbuf && (gdk_pixbuf_get_width(pixbuf) > icon->largest || gdk_pixbuf_get_height(pixbuf) > icon->largest)) {
		gint width = gdk_pixbuf_get_width(pixbuf);
		gint height = gdk_pixbuf_get_height(pixbuf);
		double scale = ((double) icon->largest) / MAX(width, height);
		GdkPixbuf *new_pixbuf;

		new_pixbuf = gdk_pixbuf_scale_simple(pixbuf, width*scale, height*scale, GDK_INTERP_BILINEAR);
		g_object_unref(pixbuf);
		pixbuf = new_pixbuf;
	}

	return pixbuf;
}

static void be_icon_set_sized(gpointer key, BeIconSized *value, BeIcon *icon) {
	GdkPixbuf *pixbuf = NULL;
	gint size;

	if(icon->pixbuf) {
		gtk_icon_size_lookup(value->icon_size, NULL, &size);

		if(gdk_pixbuf_get_width(icon->pixbuf) > size || gdk_pixbuf_get_height(icon->pixbuf) > size) {
			gint width = gdk_pixbuf_get_width(icon->pixbuf);
			gint height = gdk_pixbuf_get_height(icon->pixbuf);
			double scale = ((double) icon->largest) / MAX(width, height);

			pixbuf = gdk_pixbuf_scale_simple(icon->pixbuf, width*scale, height*scale, GDK_INTERP_BILINEAR);
		} else {
			pixbuf = icon->pixbuf;
			g_object_ref(G_OBJECT(pixbuf));
		}
	}

	be_icon_sized_set(value, pixbuf);

	if(pixbuf)
		g_object_unref(G_OBJECT(pixbuf));
}
static void be_icon_set_sizes(BeIcon *icon) {
	g_hash_table_foreach(icon->sized, (GHFunc) &be_icon_set_sized, icon);
}

static void be_icon_reload_pixbuf(BeIcon *icon) {
	gchar *path = gnome_desktop_item_find_icon(NULL, icon->name, icon->largest, 0);

	if(path && (!icon->path || strcmp(path, icon->path) != 0)) {
		GdkPixbuf *pixbuf;

		g_free(icon->path);
		icon->path = path;

		pixbuf = be_icon_load_pixbuf(icon);

		if(pixbuf) {
			if(icon->pixbuf)
				g_object_unref(icon->pixbuf);
			icon->pixbuf = pixbuf;
		}
	} else if(icon->pixbuf && icon->path) {
		/*
		 * I chose AND instead of OR because if one dimension is large
		 * enough, apparently we've already loaded for a larger size.
		 */
		if(gdk_pixbuf_get_width(icon->pixbuf) < icon->largest && gdk_pixbuf_get_height(icon->pixbuf) < icon->largest) {
			GdkPixbuf *pixbuf = be_icon_load_pixbuf(icon);

			if(pixbuf) {
				g_object_unref(icon->pixbuf);
				icon->pixbuf = pixbuf;
				be_icon_set_sizes(icon);
			}
		}

		g_free(path);
	} else {
		if(icon->pixbuf) {
			g_object_unref(G_OBJECT(icon->pixbuf));
			icon->pixbuf = NULL;
		}

		g_free(path);
	}

	be_icon_set_sizes(icon);
}

static void be_icon_set_icon_theme(BeIcon *icon, GtkIconTheme *icon_theme) {
	if(icon->icon_theme != icon_theme) {
		icon->icon_theme = icon_theme;

		if(icon->icon_theme) {
			be_icon_reload_pixbuf(icon);
		}
	}
}
static GtkIconTheme *be_icon_get_icon_theme(const BeIcon *icon) {
	return icon->icon_theme;
}

static void be_icon_add_image(BeIcon *icon, GtkImage *image, GtkIconSize icon_size) {
	BeIconSized *icon_sized = g_hash_table_lookup(icon->sized, GUINT_TO_POINTER(icon_size));

	if(!icon_sized) {
		gint size;

		gtk_icon_size_lookup(icon_size, NULL, &size);

		icon_sized = be_icon_sized_new(icon_size);

		g_hash_table_insert(icon->sized, GUINT_TO_POINTER(icon_size), icon_sized);

		if(size > icon->largest) {
			icon->largest = size;

			if(icon->icon_theme) {
				be_icon_reload_pixbuf(icon);
			}
		} else {
			be_icon_set_sized(GUINT_TO_POINTER(icon_size), icon_sized, icon);
		}
	}

	be_icon_sized_add_image(icon_sized, image);
}

static BeIcon *be_icon_new(const gchar *name, GtkIconTheme *icon_theme) {
	BeIcon *icon = g_new0(BeIcon, 1);

	icon->name  = g_strdup(name);
	icon->sized = g_hash_table_new_full(g_direct_hash, g_direct_equal, NULL, (GDestroyNotify) &be_icon_sized_free);

	be_icon_set_icon_theme(icon, icon_theme);

	return icon;
}
static void be_icon_free(BeIcon *icon) {
	g_free(icon->name);
	g_free(icon->path);

	if(icon->pixbuf)
		g_object_unref(icon->pixbuf);

	g_hash_table_destroy(icon->sized);

	g_free(icon);
}

/* BeIconLoader */
enum {
	BE_ICON_LOADER_PROPERTY_ICON_THEME = 1
};

static gboolean be_icon_loader_idle(BeIconLoader *icon_loader) {
	BeIcon *value = g_queue_pop_tail(icon_loader->dirty);

	be_icon_set_icon_theme(value, icon_loader->icon_theme);

	if(g_queue_is_empty(icon_loader->dirty)) {
		icon_loader->idle = 0;
		return FALSE;
	}

	return TRUE;
}

static void be_icon_loader_append_dirty(BeIconLoader *icon_loader, BeIcon *value) {
	if(!g_queue_find(icon_loader->dirty, value)) {
		g_queue_push_head(icon_loader->dirty, value);

		if(icon_loader->idle == 0)
			icon_loader->idle = g_idle_add((GSourceFunc) &be_icon_loader_idle, icon_loader);
	}
}
static void be_icon_loader_remove_dirty(BeIconLoader *icon_loader, BeIcon *value) {
	g_queue_remove(icon_loader->dirty, value);
}

static void be_icon_loader_icon_dirty(gpointer key, BeIcon *value, BeIconLoader *icon_loader) {
	be_icon_set_icon_theme(value, NULL);
	be_icon_loader_append_dirty(icon_loader, value);
}
static void be_icon_loader_icon_theme_changed_virtual(BeIconLoader *icon_loader) {
	if(icon_loader->icon_theme)
		g_hash_table_foreach(icon_loader->icons, (GHFunc) &be_icon_loader_icon_dirty, icon_loader);
}

static BeIcon *be_icon_loader_append(BeIconLoader *icon_loader, const gchar *name, GtkIconSize icon_size) {
	BeIcon *icon = be_icon_new(name, NULL);

	g_hash_table_insert(icon_loader->icons, g_strdup(name), icon);
	be_icon_loader_append_dirty(icon_loader, icon);

	return icon;
}

static void be_icon_loader_icon_theme_changed_callback(BeIconLoader *icon_loader, GtkIconTheme *icon_theme) {
	be_icon_loader_icon_theme_changed(icon_loader);
}

static void be_icon_loader_instance_init(GTypeInstance *instance, gpointer g_class) {
	BeIconLoader *self = (BeIconLoader *) instance;

	/* Do stuff */
	self->icons = g_hash_table_new_full(g_str_hash, g_str_equal, g_free, (GDestroyNotify) be_icon_free);
	self->dirty = g_queue_new();
	self->idle  = 0;
}

static void be_icon_loader_set_property(GObject *object, guint property_id, const GValue *value, GParamSpec *pspec) {
	BeIconLoader *self = BE_ICON_LOADER(object);

	switch(property_id) {
		case BE_ICON_LOADER_PROPERTY_ICON_THEME:
			be_icon_loader_set_icon_theme(self, GTK_ICON_THEME(g_value_get_object(value)));
			break;
		default:
			g_assert_not_reached();
			break;
	}
}

static void be_icon_loader_get_property(GObject *object, guint property_id, GValue *value, GParamSpec *pspec) {
	BeIconLoader *self = BE_ICON_LOADER(object);

	switch(property_id) {
		case BE_ICON_LOADER_PROPERTY_ICON_THEME:
			g_value_set_object(value, G_OBJECT(self->icon_theme));
			break;
		default:
			g_assert_not_reached();
			break;
	}
}

static GObject *be_icon_loader_constructor(GType type, guint n_construct_properties, GObjectConstructParam *construct_properties) {
	GObject *obj;

	{ /* Invoke parent constructor. */
		BeIconLoaderClass *klass        = BE_ICON_LOADER_CLASS(g_type_class_peek(BE_ICON_LOADER_TYPE));
		GObjectClass      *parent_class = G_OBJECT_CLASS(g_type_class_peek_parent(klass));

		obj = parent_class->constructor(type, n_construct_properties, construct_properties);
	}

	/* Do stuff */

	return obj;
}

static void be_icon_loader_dispose(GObject *object) {
	BeIconLoader      *self         = BE_ICON_LOADER(object);
	BeIconLoaderClass *klass        = BE_ICON_LOADER_CLASS(g_type_class_peek(BE_ICON_LOADER_TYPE));
	GObjectClass      *parent_class = G_OBJECT_CLASS(g_type_class_peek_parent(klass));

	/* Do stuff */
	if(self->icon_theme) {
		g_signal_handlers_disconnect_by_func(self->icon_theme, G_CALLBACK(&be_icon_loader_icon_theme_changed_callback), self);
		g_object_unref(self->icon_theme);
		self->icon_theme = NULL;
	}

	/* Invoke parent dispose */
	parent_class->dispose(object);
}

static void be_icon_loader_finalize(GObject *object) {
	BeIconLoader      *self         = BE_ICON_LOADER(object);
	BeIconLoaderClass *klass        = BE_ICON_LOADER_CLASS(g_type_class_peek(BE_ICON_LOADER_TYPE));
	GObjectClass      *parent_class = G_OBJECT_CLASS(g_type_class_peek_parent(klass));

	/* Do stuff */
	g_hash_table_destroy(self->icons);
	g_queue_free(self->dirty);
	if(self->idle != 0)
		g_idle_remove_by_data(self);

	/* Invoke parent finalize */
	parent_class->finalize(object);
}

static void be_icon_loader_class_init(gpointer g_class, gpointer g_class_data) {
	GObjectClass      *gobject_class = G_OBJECT_CLASS(g_class);
	BeIconLoaderClass *klass         = BE_ICON_LOADER_CLASS(g_class);
	GParamSpec        *pspec;

	gobject_class->constructor  = &be_icon_loader_constructor;
	gobject_class->dispose      = &be_icon_loader_dispose;
	gobject_class->finalize     = &be_icon_loader_finalize;
	gobject_class->set_property = &be_icon_loader_set_property;
	gobject_class->get_property = &be_icon_loader_get_property;

	klass->icon_theme_changed   = &be_icon_loader_icon_theme_changed_virtual;

	pspec = g_param_spec_object("icon-theme",
		"Icon theme",
		"Icon theme to use for looking up icons to load and to monitor for changes",
		GTK_TYPE_ICON_THEME,
		G_PARAM_READWRITE);
	g_object_class_install_property(gobject_class,
		BE_ICON_LOADER_PROPERTY_ICON_THEME,
		pspec);
}

GType be_icon_loader_get_type(void) {
	static GType type = 0;

	if(type == 0) {
		static const GTypeInfo info = {
			sizeof(BeIconLoaderClass),     /* class_size */
			NULL,                          /* base_init */
			NULL,                          /* base_finalize */
			&be_icon_loader_class_init,    /* class_init */
			NULL,                          /* class_destroy */
			NULL,                          /* class_data */
			sizeof(BeIconLoader),          /* instance_size */
			0,                             /* n_preallocs */
			&be_icon_loader_instance_init, /* instance_init */
			NULL                           /* value_table */
		};

		type = g_type_register_static(G_TYPE_OBJECT, "BeIconLoaderType", &info, 0);
	}

	return type;
}

BeIconLoader *be_icon_loader_new(GtkIconTheme *icon_theme) {
	BeIconLoader *icon_loader;

	if(!icon_theme)
		icon_theme = gtk_icon_theme_get_default();

	icon_loader = BE_ICON_LOADER(g_object_new(BE_ICON_LOADER_TYPE,
		"icon-theme", icon_theme,
		NULL));

	return icon_loader;
}

void be_icon_loader_set_icon_theme(BeIconLoader *icon_loader, GtkIconTheme *icon_theme) {
	if(icon_loader->icon_theme) {
		g_signal_handlers_disconnect_by_func(icon_loader->icon_theme, G_CALLBACK(&be_icon_loader_icon_theme_changed_callback), icon_loader);
		g_object_unref(icon_loader->icon_theme);
	}

	icon_loader->icon_theme = icon_theme;

	if(icon_loader->icon_theme) {
		g_object_ref(icon_loader->icon_theme);
		g_signal_connect_object(icon_loader->icon_theme, "changed", G_CALLBACK(&be_icon_loader_icon_theme_changed_callback), icon_loader, G_CONNECT_SWAPPED);
	}

	g_object_notify(G_OBJECT(icon_loader), "icon-theme");
	be_icon_loader_icon_theme_changed(icon_loader);
}
GtkIconTheme *be_icon_loader_get_icon_theme(const BeIconLoader *icon_loader) {
	return icon_loader->icon_theme;
}

void be_icon_loader_icon_theme_changed(BeIconLoader *icon_loader) {
	BE_ICON_LOADER_GET_CLASS(icon_loader)->icon_theme_changed(icon_loader);
}

void be_icon_loader_add_image(BeIconLoader *icon_loader, GtkImage *image, const gchar *name, GtkIconSize icon_size) {
	BeIcon *icon = g_hash_table_lookup(icon_loader->icons, name);

	if(!icon)
		icon = be_icon_loader_append(icon_loader, name, icon_size);

	be_icon_add_image(icon, image, icon_size);
}
GtkWidget *be_icon_loader_image_new(BeIconLoader *icon_loader, const gchar *name, GtkIconSize icon_size) {
	GtkWidget *image = gtk_image_new();

	be_icon_loader_add_image(icon_loader, GTK_IMAGE(image), name, icon_size);

	return image;
}
