#include "config.h"

#include "window-list.h"

#include <libgnome/gnome-i18n.h>
#include <libgnome/gnome-desktop-item.h>
#include <libgnomeui/libgnomeui.h>
#include <libwnck/window-action-menu.h>

#ifdef HAVE_STARTUP_NOTIFICATION
	#include <gdk/gdkx.h>
#endif

/* Thanks to libwnck for information used to implement this file */

static void be_window_list_append_window(BeWindowList *window_list, WnckWindow *window);
static void be_window_list_remove_window(BeWindowList *window_list, WnckWindow *window);

static void be_connect_while_alive(gpointer object, const gchar *signal, GCallback func, gpointer func_data, GObject *alive_object) {
	GClosure *closure;

	closure = g_cclosure_new(func, func_data, NULL);
	g_object_watch_closure(alive_object, closure);
	g_signal_connect_closure(object, signal, closure, FALSE);
}

static gint be_window_list_get_screen_number(BeWindowList *window_list) {
	GtkWidget *widget = GTK_WIDGET(window_list);
	GdkScreen *screen = gtk_widget_has_screen(widget) ? gtk_widget_get_screen(widget) : gdk_screen_get_default();

	return gdk_screen_get_number(screen);
}
static WnckScreen *be_window_list_get_screen(BeWindowList *window_list) {
	GtkWidget  *widget = GTK_WIDGET(window_list);
	WnckScreen *screen = gtk_widget_has_screen(widget) ?
		wnck_screen_get(gdk_screen_get_number(gtk_widget_get_screen(widget))) :
		wnck_screen_get_default();

	return screen;
}
static WnckWorkspace *be_window_list_get_active_workspace(BeWindowList *window_list) {
	return wnck_screen_get_active_workspace(be_window_list_get_screen(window_list));
}
static gboolean be_window_list_should_include_window(BeWindowList *window_list, WnckWindow *window) {
	WnckWorkspace *workspace = be_window_list_get_active_workspace(window_list);

	if(wnck_window_get_state(window) & WNCK_WINDOW_STATE_SKIP_TASKLIST)
		return FALSE;

	if(workspace) {
		if(!wnck_window_is_pinned(window) && workspace != wnck_window_get_workspace(window))
			return FALSE;

		if(wnck_workspace_is_virtual(workspace))
			return FALSE;

		if(!wnck_window_is_in_viewport(window, workspace))
			return FALSE;
	}

	return TRUE;
}

static GdkPixbuf *be_window_list_no_icon_new(void) {
	GdkPixbuf *pixbuf;
	gchar *default_icon_path;

	/* FIXME */
	default_icon_path = g_strdup("/usr/share/be-applet/default_icon.png");
	/* gnome_program_locate_file(NULL, GNOME_FILE_DOMAIN_APP_DATADIR, "default_icon.png", TRUE, NULL); */

	pixbuf = gdk_pixbuf_new_from_file(default_icon_path, NULL);

	g_free(default_icon_path);

	return pixbuf;
}

enum {
	BE_WINDOW_LIST_WINDOW_TEXT_DEPENDENT_STATES = (WNCK_WINDOW_STATE_MINIMIZED | WNCK_WINDOW_STATE_SHADED)
};

static gchar *be_window_list_window_get_text(BeWindowList *window_list, WnckWindow *window) {
	const gchar *name     = wnck_window_get_name(window);
	gchar       *ret_name = NULL;

	if(wnck_window_is_minimized(window) || wnck_window_is_shaded(window)) {
		ret_name = g_strdup_printf("[%s]", name);
	}

	if(!ret_name)
		ret_name = g_strdup(name);
	
	return ret_name;
}
static GdkPixbuf *be_window_list_window_get_icon(BeWindowList *window_list, WnckWindow *window) {
	GdkPixbuf *pixbuf = NULL;
	gint       width, height;
	gint       icon_size;

	gtk_icon_size_lookup(GTK_ICON_SIZE_MENU, NULL, &icon_size);

	if(window)
		pixbuf = wnck_window_get_icon(window);

	if(!pixbuf)
		pixbuf = window_list->no_icon;

	g_object_ref(pixbuf);

	width  = gdk_pixbuf_get_width(pixbuf);
	height = gdk_pixbuf_get_height(pixbuf);

	if(width > icon_size || height > icon_size) {
		double scale = ((double) icon_size) / MAX(width, height);
		GdkPixbuf *new_pixbuf;

		new_pixbuf = gdk_pixbuf_scale_simple(pixbuf, width * scale, height * scale, GDK_INTERP_BILINEAR);
		g_object_unref(pixbuf);
		pixbuf = new_pixbuf;
	}

	return pixbuf;
}

static void be_window_list_window_name_changed(WnckWindow *window, BeWindowList *window_list) {
	GtkWidget *button = g_hash_table_lookup(window_list->windows, window);

	if(button) {
		GtkWidget *label = g_object_get_data(G_OBJECT(button), "WindowList:Text");
		gchar     *text  = be_window_list_window_get_text(window_list, window);

		gtk_label_set_text(GTK_LABEL(label), text);
		gtk_tooltips_set_tip(window_list->tooltips, button, text, NULL);

		g_free(text);
	}
}
static void be_window_list_window_icon_changed(WnckWindow *window, BeWindowList *window_list) {
	GtkWidget *button = g_hash_table_lookup(window_list->windows, window);

	if(button) {
		GtkWidget *image  = g_object_get_data(G_OBJECT(button), "WindowList:Icon");
		GdkPixbuf *pixbuf = be_window_list_window_get_icon(window_list, window);

		gtk_image_set_from_pixbuf(GTK_IMAGE(image), pixbuf);

		g_object_unref(pixbuf);
	}
}
static void be_window_list_window_state_changed(WnckWindow *window, WnckWindowActions changed_mask, WnckWindowActions new_actions, BeWindowList *window_list) {
	GtkWidget *button = g_hash_table_lookup(window_list->windows, window);

	if(button) {
		if(!be_window_list_should_include_window(window_list, window))
			be_window_list_remove_window(window_list, window);
		else {
			if(changed_mask & BE_WINDOW_LIST_WINDOW_TEXT_DEPENDENT_STATES) {
				GtkWidget *label = g_object_get_data(G_OBJECT(button), "WindowList:Text");
				gchar     *text  = be_window_list_window_get_text(window_list, window);

				gtk_label_set_text(GTK_LABEL(label), text);

				g_free(text);
			}
		}
	}
}
static void be_window_list_window_workspace_changed(WnckWindow *window, BeWindowList *window_list) {
	GtkWidget *button = g_hash_table_lookup(window_list->windows, window);

	if(button) {
		if(!be_window_list_should_include_window(window_list, window))
			be_window_list_remove_window(window_list, window);
	}
}

static void be_window_list_window_widget_position_menu(GtkMenu *menu, gint *x, gint *y, gboolean *push_in, GtkWidget *widget) {
	GtkRequisition requisition;
	gint menu_xpos;
	gint menu_ypos;

	gtk_widget_size_request(GTK_WIDGET(menu), &requisition);

	gdk_window_get_origin(widget->window, &menu_xpos, &menu_ypos);

	menu_xpos += widget->allocation.x;
	menu_ypos += widget->allocation.y;

	if (menu_ypos > gdk_screen_height() / 2)
		menu_ypos -= requisition.height;
	else
		menu_ypos += widget->allocation.height;

	*x = menu_xpos;
	*y = menu_ypos;
	*push_in = TRUE;
}
static void be_window_list_window_widget_menu_selection_done(GtkMenuShell *menu_shell, gpointer user_data) {
	gtk_widget_destroy(GTK_WIDGET(menu_shell));
}
static gboolean be_window_list_window_widget_button_press_event(GtkWidget *widget, GdkEventButton *event, BeWindowList *window_list) {
	WnckWindow *window      = g_object_get_data(G_OBJECT(widget), "WindowList:Window");
	GtkWidget  *action_menu;

	switch(event->button) {
		case 1:
			if(wnck_window_is_active(window))
				window_list->was_active = widget;
			else
				window_list->was_active = NULL;
			break;
		case 3:
			action_menu = wnck_create_window_action_menu(window);

			g_signal_connect(action_menu, "selection_done", G_CALLBACK(&be_window_list_window_widget_menu_selection_done), NULL);

			gtk_menu_popup(GTK_MENU(action_menu),
				NULL, NULL,
				(GtkMenuPositionFunc) &be_window_list_window_widget_position_menu, widget,
				3, event->time);
			return TRUE;
	}

	return FALSE;
}
static void be_window_list_window_widget_toggled(GtkToggleButton *button, BeWindowList *window_list) {
	WnckWindow *window = g_object_get_data(G_OBJECT(button), "WindowList:Window");

	if(!window_list->frozen) {
		if(window_list->was_active == GTK_WIDGET(button)) {
			window_list->frozen = TRUE;

			gtk_toggle_button_set_active(button, FALSE);

			window_list->frozen = FALSE;

			wnck_window_minimize(window);
		} else
			wnck_window_activate(window);
	}
}

static GtkWidget *be_window_list_window_widget_new(BeWindowList *window_list, GdkPixbuf *pixbuf, const gchar *text) {
	GtkWidget *button = gtk_toggle_button_new();
	GtkWidget *table  = gtk_table_new(1, 2, FALSE);
	GtkWidget *widget;

	if(pixbuf)
		widget = gtk_image_new_from_pixbuf(pixbuf);
	else
		widget = gtk_image_new();
	
	gtk_table_attach(GTK_TABLE(table), widget,
		0, 1,
		0, 1,
		0, GTK_EXPAND,
		0, 0);
	g_object_set_data(G_OBJECT(button), "WindowList:Icon", GTK_IMAGE(widget));

	widget = gtk_label_new(text);
	gtk_table_attach(GTK_TABLE(table), widget,
		1, 2,
		0, 1,
		GTK_FILL, GTK_EXPAND,
		4, 0);
	g_object_set_data(G_OBJECT(button), "WindowList:Text", GTK_LABEL(widget));

	gtk_tooltips_set_tip(window_list->tooltips, button, text, NULL);

	gtk_container_add(GTK_CONTAINER(button), table);
	gtk_widget_show_all(table);

	return button;
}

static GtkWidget *be_window_list_window_new(BeWindowList *window_list, WnckWindow *window) {
	GtkWidget *button = NULL;

	if(window && be_window_list_should_include_window(window_list, window)) {
		GdkPixbuf *pixbuf = be_window_list_window_get_icon(window_list, window);
		gchar     *text   = be_window_list_window_get_text(window_list, window);

		button = be_window_list_window_widget_new(window_list, pixbuf, text);

		g_object_unref(pixbuf);
		g_free(text);

		g_object_set_data(G_OBJECT(button), "WindowList:Window", window);

		if(wnck_screen_get_active_window(be_window_list_get_screen(window_list)) == window)
			gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(button), TRUE);

		be_connect_while_alive(window, "icon_changed", G_CALLBACK(&be_window_list_window_icon_changed), window_list, G_OBJECT(window_list));
		be_connect_while_alive(window, "name_changed", G_CALLBACK(&be_window_list_window_name_changed), window_list, G_OBJECT(window_list));
		be_connect_while_alive(window, "state_changed", G_CALLBACK(&be_window_list_window_state_changed), window_list, G_OBJECT(window_list));
		be_connect_while_alive(window, "workspace_changed", G_CALLBACK(&be_window_list_window_workspace_changed), window_list, G_OBJECT(window_list));

		g_signal_connect(button, "button_press_event", G_CALLBACK(&be_window_list_window_widget_button_press_event), window_list);
		g_signal_connect(button, "toggled", G_CALLBACK(&be_window_list_window_widget_toggled), window_list);
	}

	return button;
}

static void be_window_list_append_window(BeWindowList *window_list, WnckWindow *window) {
	GtkWidget *button = be_window_list_window_new(window_list, window);

	if(button) {
		gtk_widget_show(button);

		gtk_box_pack_start(GTK_BOX(window_list), button, FALSE, FALSE, 0);

		g_hash_table_insert(window_list->windows, window, button);
	}
}
static void be_window_list_remove_window(BeWindowList *window_list, WnckWindow *window) {
	GtkWidget *button = g_hash_table_lookup(window_list->windows, window);

	if(button) {
		if(window_list->was_active == g_object_get_data(G_OBJECT(button), "WindowList:Window"))
			window_list->was_active = NULL;

		gtk_container_remove(GTK_CONTAINER(window_list), button);

		g_hash_table_remove(window_list->windows, window);
	}
}

static gboolean be_window_list_clear_windows_hrfunc(gpointer key, gpointer value, gpointer user_data) {
	BeWindowList *window_list = user_data;

	gtk_container_remove(GTK_CONTAINER(window_list), GTK_WIDGET(value));

	return TRUE;
}
static void be_window_list_clear_windows(BeWindowList *window_list) {
	g_hash_table_foreach_remove(window_list->windows, &be_window_list_clear_windows_hrfunc, window_list);
	window_list->active     =
	window_list->was_active = NULL;
}
static void be_window_list_fill_windows(BeWindowList *window_list) {
	WnckScreen *screen = be_window_list_get_screen(window_list);
	GList      *cur, *windows;

	windows = wnck_screen_get_windows(screen);

	if(windows)
		for(cur = windows; cur; cur = g_list_next(cur))
			be_window_list_append_window(window_list, cur->data);
}

static void be_window_list_refresh_windows(BeWindowList *window_list) {
	be_window_list_clear_windows(window_list);
	be_window_list_fill_windows(window_list);
}

#if HAVE_STARTUP_NOTIFICATION
	static GdkPixbuf *be_window_list_startup_sequence_get_icon(BeWindowList *window_list, SnStartupSequence *startup_seq) {
		const gchar *icon   = startup_seq ? sn_startup_sequence_get_icon_name(startup_seq) : NULL;
		GdkPixbuf   *pixbuf = NULL;
		gint         width, height;
		gint         icon_size;

		gtk_icon_size_lookup(GTK_ICON_SIZE_MENU, NULL, &icon_size);

		if(icon) {
			gchar *path = gnome_desktop_item_find_icon(NULL, icon, icon_size, 0);

			pixbuf = gdk_pixbuf_new_from_file_at_size(path, icon_size, icon_size, NULL);

			g_free(path);
		}

		if(!pixbuf) {
			pixbuf = window_list->no_icon;
			g_object_ref(pixbuf);
		}

		width  = gdk_pixbuf_get_width(pixbuf);
		height = gdk_pixbuf_get_height(pixbuf);

		if(width > icon_size || height > icon_size) {
			double scale = ((double) icon_size) / MAX(width, height);
			GdkPixbuf *new_pixbuf;

			new_pixbuf = gdk_pixbuf_scale_simple(pixbuf, width * scale, height * scale, GDK_INTERP_BILINEAR);
			g_object_unref(pixbuf);
			pixbuf = new_pixbuf;
		}

		return pixbuf;
	}

	static GtkWidget *be_window_list_startup_sequence_widget_new(BeWindowList *window_list, GdkPixbuf *pixbuf, const gchar *text) {
		GtkWidget *button = gtk_button_new();
		GtkWidget *table  = gtk_table_new(1, 2, FALSE);
		GtkWidget *widget;

		if(pixbuf)
			widget = gtk_image_new_from_pixbuf(pixbuf);
		else
			widget = gtk_image_new();
		
		gtk_table_attach(GTK_TABLE(table), widget,
			0, 1,
			0, 1,
			0, GTK_EXPAND,
			0, 0);

		widget = gtk_label_new(text);
		gtk_table_attach(GTK_TABLE(table), widget,
			1, 2,
			0, 1,
			GTK_FILL, GTK_EXPAND,
			4, 0);

		gtk_tooltips_set_tip(window_list->tooltips, button, text, NULL);

		gtk_container_add(GTK_CONTAINER(button), table);
		gtk_widget_show_all(table);

		return button;
	}

	static GtkWidget *be_window_list_startup_sequence_new(BeWindowList *window_list, SnStartupSequence *startup_seq) {
		GtkWidget   *widget = NULL;
		GdkPixbuf   *pixbuf = be_window_list_startup_sequence_get_icon(window_list, startup_seq);
		const gchar *text   = sn_startup_sequence_get_name(startup_seq);

		widget = be_window_list_startup_sequence_widget_new(window_list, pixbuf, text);

		return widget;
	}

	static void be_window_list_append_startup_sequence(BeWindowList *window_list, SnStartupSequence *startup_seq) {
		GtkWidget *button = be_window_list_startup_sequence_new(window_list, startup_seq);

		if(button) {
			gtk_widget_show(button);

			gtk_box_pack_start(GTK_BOX(window_list), button, FALSE, FALSE, 0);

			g_hash_table_insert(window_list->startup_seqs, startup_seq, button);
		}
	}
	static void be_window_list_remove_startup_sequence(BeWindowList *window_list, SnStartupSequence *startup_seq) {
		GtkWidget *button = g_hash_table_lookup(window_list->startup_seqs, startup_seq);

		if(button) {
			gtk_container_remove(GTK_CONTAINER(window_list), button);

			g_hash_table_remove(window_list->startup_seqs, startup_seq);
		}
	}
#endif

static void be_window_list_active_window_changed(WnckScreen *screen, BeWindowList *window_list) {
	WnckWindow *window = wnck_screen_get_active_window(screen);

	if(!window_list->frozen) {
		GtkWidget *button = g_hash_table_lookup(window_list->windows, window);

		if(button != window_list->active) {
			window_list->frozen = TRUE;

			if(window_list->active)
				gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(window_list->active), FALSE);

			window_list->active = button;

			if(window_list->active)
				gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(window_list->active), TRUE);

			window_list->frozen = FALSE;
		}
	}
}
static void be_window_list_window_opened(WnckScreen *screen, WnckWindow *window, BeWindowList *window_list) {
	be_window_list_append_window(window_list, window);
}
static void be_window_list_window_closed(WnckScreen *screen, WnckWindow *window, BeWindowList *window_list) {
	be_window_list_remove_window(window_list, window);
}
static void be_window_list_active_workspace_changed(WnckWindow *screen, BeWindowList *window_list) {
	be_window_list_refresh_windows(window_list);
}

static void be_window_list_instance_init(GTypeInstance *instance, gpointer g_class) {
	BeWindowList *self = (BeWindowList *) instance;
	WnckScreen   *screen = be_window_list_get_screen(self);

	/* Do stuff */
	self->tooltips   = gtk_tooltips_new();
	self->windows    = g_hash_table_new(g_direct_hash, g_direct_equal);
	self->no_icon    = be_window_list_no_icon_new();
	self->active     =
	self->was_active = NULL;
	self->frozen     = FALSE;

	#if HAVE_STARTUP_NOTIFICATION
		self->sn_monitor   = NULL;
		self->startup_seqs = g_hash_table_new(g_direct_hash, g_direct_equal);
	#endif

	g_object_ref(G_OBJECT(self->tooltips));
	gtk_object_sink(GTK_OBJECT(self->tooltips));

	be_window_list_fill_windows(self);

	be_connect_while_alive(screen, "active_window_changed", G_CALLBACK(&be_window_list_active_window_changed), self, G_OBJECT(self));
	be_connect_while_alive(screen, "window_opened", G_CALLBACK(&be_window_list_window_opened), self, G_OBJECT(self));
	be_connect_while_alive(screen, "window_closed", G_CALLBACK(&be_window_list_window_closed), self, G_OBJECT(self));
	be_connect_while_alive(screen, "active_workspace_changed", G_CALLBACK(&be_window_list_active_workspace_changed), self, G_OBJECT(self));
}

static GObject *be_window_list_constructor(GType type, guint n_construct_properties, GObjectConstructParam *construct_properties) {
	GObject *obj;

	{ /* Invoke parent constructor. */
		BeWindowListClass *klass        = BE_WINDOW_LIST_CLASS(g_type_class_peek(BE_WINDOW_LIST_TYPE));
		GObjectClass      *parent_class = G_OBJECT_CLASS(g_type_class_peek_parent(klass));

		obj = parent_class->constructor(type, n_construct_properties, construct_properties);
	}

	/* Do stuff */

	return obj;
}

#if 0
static void be_window_list_dispose(GObject *object) {
	BeWindowListClass *klass        = BE_WINDOW_LIST_CLASS(g_type_class_peek(BE_WINDOW_LIST_TYPE));
	GObjectClass      *parent_class = G_OBJECT_CLASS(g_type_class_peek_parent(klass));

	/* Do stuff */

	/* Invoke parent dispose */
	parent_class->dispose(object);
}
#endif

static void be_window_list_finalize(GObject *object) {
	BeWindowList      *self         = BE_WINDOW_LIST(object);
	BeWindowListClass *klass        = BE_WINDOW_LIST_CLASS(g_type_class_peek(BE_WINDOW_LIST_TYPE));
	GObjectClass      *parent_class = G_OBJECT_CLASS(g_type_class_peek_parent(klass));

	/* Do stuff */
	g_object_unref(self->tooltips);
	g_hash_table_destroy(self->windows);
	g_object_unref(self->no_icon);

	#if HAVE_STARTUP_NOTIFICATION
		if(self->sn_monitor)
			sn_monitor_context_unref(self->sn_monitor);
		g_hash_table_destroy(self->startup_seqs);
	#endif

	/* Invoke parent finalize */
	parent_class->finalize(object);
}

#ifdef HAVE_STARTUP_NOTIFICATION
	static void be_window_list_sn_display_error_trap_push(SnDisplay *display, Display *xdisplay) {
		gdk_error_trap_push();
	}
	static void be_window_list_sn_display_error_trap_pop(SnDisplay *display, Display *xdisplay) {
		gdk_error_trap_pop();
	}

	static SnDisplay *be_window_list_sn_display_new(BeWindowList *window_list) {
		SnDisplay *display = sn_display_new(
			gdk_x11_drawable_get_xdisplay(GDK_DRAWABLE(gtk_widget_get_parent_window(GTK_WIDGET(window_list)))),
			&be_window_list_sn_display_error_trap_push,
			&be_window_list_sn_display_error_trap_pop);

		return display;
	}

	static void be_window_list_sn_monitor_event(SnMonitorEvent *event, void *user_data) {
		BeWindowList      *window_list = BE_WINDOW_LIST(user_data);
		SnStartupSequence *startup_seq = sn_monitor_event_get_startup_sequence(event);

		switch(sn_monitor_event_get_type(event)) {
			case SN_MONITOR_EVENT_INITIATED:
				be_window_list_append_startup_sequence(window_list, startup_seq);
				break;
			case SN_MONITOR_EVENT_COMPLETED:
				be_window_list_remove_startup_sequence(window_list, startup_seq);
				break;
			case SN_MONITOR_EVENT_CHANGED:
				break;
			case SN_MONITOR_EVENT_CANCELED:
				be_window_list_remove_startup_sequence(window_list, startup_seq);
				break;
		}
	}

#if 0
	  if (tasklist->priv->startup_sequences == NULL &&
	      tasklist->priv->startup_sequence_timeout != 0)
	    {
	      g_source_remove (tasklist->priv->startup_sequence_timeout);
	      tasklist->priv->startup_sequence_timeout = 0;
	    }
	}
#endif

	static void be_window_list_realize(GtkWidget *widget) {
		BeWindowList      *self         = BE_WINDOW_LIST(widget);
		BeWindowListClass *klass        = BE_WINDOW_LIST_CLASS(g_type_class_peek(BE_WINDOW_LIST_TYPE));
		GtkWidgetClass    *parent_class = GTK_WIDGET_CLASS(g_type_class_peek_parent(klass));

		/* Invoke parent realize */
		parent_class->realize(widget);

		if(!self->sn_monitor) {
			SnDisplay *sn_display = be_window_list_sn_display_new(self);
			gint       screen     = be_window_list_get_screen_number(self);

			/* Make sure libwnck has setup the root window mask for us */
			wnck_screen_get(screen);

			self->sn_monitor = sn_monitor_context_new(
				sn_display,
				screen,
				&be_window_list_sn_monitor_event,
				self,
				NULL);

			sn_display_unref(sn_display);
		}
	}
	static void be_window_list_unrealize(GtkWidget *widget) {
		BeWindowList      *self         = BE_WINDOW_LIST(widget);
		BeWindowListClass *klass        = BE_WINDOW_LIST_CLASS(g_type_class_peek(BE_WINDOW_LIST_TYPE));
		GtkWidgetClass    *parent_class = GTK_WIDGET_CLASS(g_type_class_peek_parent(klass));

		if(self->sn_monitor) {
			sn_monitor_context_unref(self->sn_monitor);
			self->sn_monitor = NULL;
		}

		/* Invoke parent unrealize */
		parent_class->unrealize(widget);
	}
#endif

static void be_window_list_class_init(gpointer g_class, gpointer g_class_data) {
	GObjectClass      *gobject_class = G_OBJECT_CLASS(g_class);
	GtkWidgetClass    *widget_class  = GTK_WIDGET_CLASS(g_class);
/*	BeWindowListClass *klass         = BE_WINDOW_LIST_CLASS(g_class); */

	gobject_class->constructor = &be_window_list_constructor;
/*	gobject_class->dispose     = &be_window_list_dispose; */
	gobject_class->finalize    = &be_window_list_finalize;

	#if HAVE_STARTUP_NOTIFICATION
		widget_class->realize   = &be_window_list_realize;
		widget_class->unrealize = &be_window_list_unrealize;
	#endif
}

GType be_window_list_get_type(void) {
	static GType type = 0;

	if(type == 0) {
		static const GTypeInfo info = {
			sizeof(BeWindowListClass),     /* class_size */
			NULL,                          /* base_init */
			NULL,                          /* base_finalize */
			&be_window_list_class_init,    /* class_init */
			NULL,                          /* class_destroy */
			NULL,                          /* class_data */
			sizeof(BeWindowList),          /* instance_size */
			0,                             /* n_preallocs */
			&be_window_list_instance_init, /* instance_init */
			NULL                           /* value_table */
		};

		type = g_type_register_static(GTK_TYPE_VBOX, "BeWindowListType", &info, 0);
	}

	return type;
}

GtkWidget *be_window_list_new(void) {
	GtkWidget *window_list = gtk_widget_new(BE_WINDOW_LIST_TYPE,
		"homogeneous", TRUE,
		"spacing", 0,
		NULL);

	return window_list;
}
